program main
!-----------------------------------------------------------------------
!     one-dimensional relativistic magnetohydrodynamic code
!     without the normal magnetic field (B_x=0)
!-----------------------------------------------------------------------
!     2009/08/01  S. Zenitani  1D HLL/HLLD solver
!     2010/04/26  S. Zenitani  a bug fixed (sound speed in HLL flux)
!-----------------------------------------------------------------------
  implicit none
  integer, parameter :: ix = 6402
  integer, parameter :: margin = 1
  integer, parameter :: loop_max = 30000
  real(8), parameter :: cfl   = 0.8d0 / 2  ! Two Riemann fans (R+L)
  real(8), parameter :: tend  = 0.4d0
  real(8), parameter :: dtout = 0.1d0 ! output interval
  integer, parameter :: lm_type = 2   ! limiter (0: flat, 1: minmod, 2: MC)
  logical, parameter :: hlld = .true. ! HLLD or HLL
  integer :: i, j, k, kout
  real(8) :: t, dt, tout
  character*256 :: filename
!-----------------------------------------------------------------------
  real(8) :: x(ix),dx
! conserved variables (U)
  real(8) :: Ud(ix), Um(ix,3), Ue(ix)
  real(8) :: Ub(ix,3)
! conserved variables (U*)
  real(8) :: U1d(ix), U1m(ix,3), U1e(ix)
  real(8) :: U1b(ix,3)
! primitive variables (V)
  real(8) :: Vd(ix), Vv(ix,3), Vp(ix)
! left states (VL)
  real(8) :: VLd(ix), VLv(ix,3), VLp(ix)
  real(8) :: ULb(ix,3)
! right states (VR)
  real(8) :: VRd(ix), VRv(ix,3), VRp(ix)
  real(8) :: URb(ix,3)
! numerical flux (F)
  real(8) :: Fd(ix), Fm(ix,3), Fe(ix), Fb(ix,3)
!-----------------------------------------------------------------------

  call model(Ud,Um,Ue,Ub,margin,x,dx,ix)
  dt   =  cfl * dx
  t    =  0.d0
  tout = -dt/3.d0
  kout =  0

  if ( dt .gt. dtout ) then
     write(6,*) 'error: ', dt, '>', dtout
     stop
  endif
  write(6,*) '[Params]'
  write(6,998) dt, dtout, ix
998 format ('dt:', 1p, e10.3, ' dtout:', 1p, e10.3, ' grids:', i5 )
  write(6,*) '== start =='

!-----------------------------------------------------------------------
  do k=1,loop_max

!    U ==> V (1)
!     write(6,*) 'U --> V'
     call u2v(Ud,Um,Ue,Ub,Vd,Vv,Vp,ix)
!   -----------------  
!    [ output ]
     if ( t .ge. tout ) then
        write(6,*) 'data output   t = ', t
        write(filename,990) kout
990     format ('data/field-',i3.3,'.dat')
        call output(filename,ix,t,x,Ud,Um,Ue,Ub,Vd,Vv,Vp)
        kout = kout + 1
        tout = tout + dtout
     endif
!    [ end? ]
     if ( k .eq. loop_max ) then
        write(6,*) 'max loop'
        goto 1000
     endif
     if ( t .ge. tend ) then
        goto 1000
     endif
!   -----------------  
!    V ==> VR, VL
     call limiter(Vd,VLd,VRd,ix,lm_type)
     call limiter(Vv(1,1),VLv(1,1),VRv(1,1),ix,lm_type)
     call limiter(Vv(1,2),VLv(1,2),VRv(1,2),ix,lm_type)
     call limiter(Vv(1,3),VLv(1,3),VRv(1,3),ix,lm_type)
     call limiter(Vp,VLp,VRp,ix,lm_type)
     call limiter(Ub(1,1),ULb(1,1),URb(1,1),ix,lm_type)
     call limiter(Ub(1,2),ULb(1,2),URb(1,2),ix,lm_type)
     call limiter(Ub(1,3),ULb(1,3),URb(1,3),ix,lm_type)
!     write(6,*) Vv(16,1), VLv(16,1), VRv(16,1)
!    VR, VL ==> Riemann Flux
     call hll(Fd,Fm,Fe,Fb,VLd,VLv,VLp,ULb,VRd,VRv,VRp,URb,ix,hlld)
!    U* = U + (dt/dx) (F-F)
     call rk21(Ud,U1d,Fd,dt,dx,ix)
     call rk21(Um(1,1),U1m(1,1),Fm(1,1),dt,dx,ix)
     call rk21(Um(1,2),U1m(1,2),Fm(1,2),dt,dx,ix)
     call rk21(Um(1,3),U1m(1,3),Fm(1,3),dt,dx,ix)
     call rk21(Ue,U1e,Fe,dt,dx,ix)
     call rk21(Ub(1,1),U1b(1,1),Fb(1,1),dt,dx,ix)
     call rk21(Ub(1,2),U1b(1,2),Fb(1,2),dt,dx,ix)
     call rk21(Ub(1,3),U1b(1,3),Fb(1,3),dt,dx,ix)
!    U* ==> V (1)
!     write(6,*) 'U* --> V'
     call u2v(U1d,U1m,U1e,U1b,Vd,Vv,Vp,ix)
!    V ==> VR, VL
     call limiter(Vd,VLd,VRd,ix,lm_type)
     call limiter(Vv(1,1),VLv(1,1),VRv(1,1),ix,lm_type)
     call limiter(Vv(1,2),VLv(1,2),VRv(1,2),ix,lm_type)
     call limiter(Vv(1,3),VLv(1,3),VRv(1,3),ix,lm_type)
     call limiter(Vp,VLp,VRp,ix,lm_type)
     call limiter(U1b(1,1),ULb(1,1),URb(1,1),ix,lm_type)
     call limiter(U1b(1,2),ULb(1,2),URb(1,2),ix,lm_type)
     call limiter(U1b(1,3),ULb(1,3),URb(1,3),ix,lm_type)
!    VR, VL ==> Riemann Flux
     call hll(Fd,Fm,Fe,Fb,VLd,VLv,VLp,ULb,VRd,VRv,VRp,URb,ix,hlld)
!    U_new = 0.5( U_old + U* + F dt )
     call rk22(Ud,U1d,Fd,dt,dx,ix)
     call rk22(Um(1,1),U1m(1,1),Fm(1,1),dt,dx,ix)
     call rk22(Um(1,2),U1m(1,2),Fm(1,2),dt,dx,ix)
     call rk22(Um(1,3),U1m(1,3),Fm(1,3),dt,dx,ix)
     call rk22(Ue,U1e,Fe,dt,dx,ix)
     call rk22(Ub(1,1),U1b(1,1),Fb(1,1),dt,dx,ix)
     call rk22(Ub(1,2),U1b(1,2),Fb(1,2),dt,dx,ix)
     call rk22(Ub(1,3),U1b(1,3),Fb(1,3),dt,dx,ix)
     t=t+dt
  enddo

1000 continue

  write(6,*) '== end =='
end program main
!-----------------------------------------------------------------------
