subroutine hll(Fd,Fm,Fe,Fb,VLd,VLv,VLp,ULb,VRd,VRv,VRp,URb,ix,hlld)
!-----------------------------------------------------------------------
!     HLL/HLLD solver
!     Ref. Mignone & Bodo (2006)
!     Ref. Mignone et al. (2009) section 3.4.1
!-----------------------------------------------------------------------
!     2009/08/01  S. Zenitani  standard HLL solver
!     2009/08/05  S. Zenitani   + simplified HLLD solver
!     2010/04/26  S. Zenitani  a bug fixed (sound speed in HLL flux)
!-----------------------------------------------------------------------
  implicit none
  real(8), parameter :: threshold = 1.0d-5
! numerical flux (F) [output]
  real(8) :: Fd(ix), Fm(ix,3), Fe(ix), Fb(ix,3)
! left states (VL) [input]
  real(8) :: VLd(ix), VLv(ix,3), VLp(ix)
  real(8) :: ULb(ix,3)
! right states (VR) [input]
  real(8) :: VRd(ix), VRv(ix,3), VRp(ix)
  real(8) :: URb(ix,3)
! left states (UL & UR; local)
  real(8) :: ULd(ix), ULm(ix,3), ULe(ix)
  real(8) :: URd(ix), URm(ix,3), URe(ix)
! numerical flux (FL & FR local)
  real(8) :: FLd(ix), FLm(ix,3), FLe(ix), FLb(ix,3)
  real(8) :: FRd(ix), FRm(ix,3), FRe(ix), FRb(ix,3)
  integer :: ix, i
  logical :: hlld

  real(8) :: v2, u0, u02, W, B2, vB
  real(8) :: aL, aR, aLR, a
  real(8) :: f1L, f1R, f2L, f2R, f3L, f3R
  real(8) :: a0L, a0R, a1L, a1R, a2L, a2R
! for hlld states
  real(8) :: pt, ehll, mxhll, a1, a0
  real(8) :: vx, vy, vz, by, bz, f1, f2
  real(8) :: Rd, Re, Rmx, Rmy, Rmz, Rby, Rbz

  call v2u(VLd,VLv,VLp,ULb,ULd,ULm,ULe,ix,1,ix-1)
  call v2f(VLd,VLv,VLp,ULb,FLd,FLm,FLe,FLb,ix,1,ix-1)
  call v2u(VRd,VRv,VRp,URb,URd,URm,URe,ix,1,ix-1)
  call v2f(VRd,VRv,VRp,URb,FRd,FRm,FRe,FRb,ix,1,ix-1)

  do i=1,ix-1

!! ========== This requires B=0 ========== !!
!    VL -> coefficients
     v2 = dot_product( VLv(i,1:3), VLv(i,1:3) )
     vB = dot_product( VLv(i,1:3), ULb(i,1:3) )
     B2 = dot_product( ULb(i,1:3), ULb(i,1:3) )
!    f1: cs^2 = (Gamma p / w) = (4p)/(3ro+12p)
!    f2: gamma^2
!    f3: Q/w
     f1L = 4*VLp(i) / ( 3*VLd(i)+12*VLp(i) )
     f2L = 1.d0 / ( 1.d0 - v2 )
     f3L = ( B2-v2*B2+vB**2-f1L*(vB**2) )/( VLd(i)+4*VLp(i) )
     a2L = f1L + f2L*(1-f1L) + f3L
     a1L = -2*f2L*VLv(i,1)*(1-f1L)
     a0L = -f1L + f2L*(VLv(i,1)**2)*(1-f1L) - f3L
!    VR -> coefficients
     v2 = dot_product( VRv(i,1:3), VRv(i,1:3) )
     vB = dot_product( VRv(i,1:3), URb(i,1:3) )
     B2 = dot_product( URb(i,1:3), URb(i,1:3) )
     f1R = 4*VRp(i) / ( 3*VRd(i)+12*VRp(i) )
     f2R = 1.d0 / ( 1.d0 - v2 )
     f3R = ( B2-v2*B2+vB**2-f1R*(vB**2) )/( VRd(i)+4*VRp(i) )
     a2R = f1R + f2R*(1-f1R) + f3R
     a1R = -2*f2R*VRv(i,1)*(1-f1R)
     a0R = -f1R + f2R*(VRv(i,1)**2)*(1-f1R) - f3R
!! ========== This requires B=0 ========== !!

!    Riemann fan speed
     aL = min( &
          (-a1L-sqrt(a1L**2-4*a2L*a0L))/(2*a2L), &
          (-a1R-sqrt(a1R**2-4*a2R*a0R))/(2*a2R) )
     aR = max( &
          (-a1L+sqrt(a1L**2-4*a2L*a0L))/(2*a2L), &
          (-a1R+sqrt(a1R**2-4*a2R*a0R))/(2*a2R) )
     aLR = aL*aR

! ========== standard HLL ==========
!    F = F(L)
     if ( aL .gt. 0 ) then
        Fd(i)   = FLd(i)
        Fm(i,1) = FLm(i,1)
        Fm(i,2) = FLm(i,2)
        Fm(i,3) = FLm(i,3)
        Fe(i)   = FLe(i)
        Fb(i,1) = FLb(i,1)
        Fb(i,2) = FLb(i,2)
        Fb(i,3) = FLb(i,3)
!    F = F(R)
     elseif ( aR .lt. 0 ) then
        Fd(i)   = FRd(i)
        Fm(i,1) = FRm(i,1)
        Fm(i,2) = FRm(i,2)
        Fm(i,3) = FRm(i,3)
        Fe(i)   = FRe(i)
        Fb(i,1) = FRb(i,1)
        Fb(i,2) = FRb(i,2)
        Fb(i,3) = FRb(i,3)
!    F = F(HLL)
     else
        a = 1.d0 / ( aR - aL )
        Fd(i)  =a*(aR*FLd(i)  -aL*FRd(i)  +aLR*(URd(i)  -ULd(i)  ))
        Fm(i,1)=a*(aR*FLm(i,1)-aL*FRm(i,1)+aLR*(URm(i,1)-ULm(i,1)))
        Fm(i,2)=a*(aR*FLm(i,2)-aL*FRm(i,2)+aLR*(URm(i,2)-ULm(i,2)))
        Fm(i,3)=a*(aR*FLm(i,3)-aL*FRm(i,3)+aLR*(URm(i,3)-ULm(i,3)))
        Fe(i)  =a*(aR*FLe(i)  -aL*FRe(i)  +aLR*(URe(i)  -ULe(i)  ))
        Fb(i,1)=a*(aR*FLb(i,1)-aL*FRb(i,1)+aLR*(URb(i,1)-ULb(i,1)))
        Fb(i,2)=a*(aR*FLb(i,2)-aL*FRb(i,2)+aLR*(URb(i,2)-ULb(i,2)))
        Fb(i,3)=a*(aR*FLb(i,3)-aL*FRb(i,3)+aLR*(URb(i,3)-ULb(i,3)))

! ========== HLLD module ==========
        if( hlld ) then
           ehll  = a*(aR*URe(i)-aL*ULe(i)+FLe(i)-FRe(i))
           mxhll = a*(aR*URm(i,1)-aL*ULm(i,1)+FLm(i,1)-FRm(i,1))
           
!       calculate covariant pressure
!       pt = p + 0.5 |b|^2
           a1 = ehll - Fm(i,1)
           a0 = mxhll * Fe(i) - Fm(i,1) * ehll
           pt = (-a1+sqrt(a1**2-4*a0)) / 2.d0

!       vx : speed of contact/tangential discon.
           vx = ( pt + aR*URm(i,1) - FRm(i,1) ) &
                / ( aR*( URe(i)+pt ) - FRe(i) )
           if ( ( vx .lt. aL).or.( vx .gt. aR )) then
              write(6,*) 'something is wrong in HLLD: ', aL, vx, aR
              stop
           endif
!          use HLLD only when |aR-vx| or |vx-aL| > threshold
           if( ( (aR-vx).gt.threshold ).and.( (vx-aL).gt.threshold ) ) then
!              write(6,*) 'HLLD'
!             HLLF flux : F_aL = FL + aL*(U_aL-UL)
!             Anyway, we are dealing with discrete numerical values
!             f2 = lambda * p + Re = E + p
              if ( vx .ge. 0 ) then
                 f1 = 1.d0 / ( aL - vx )
                 Rd  = ( aL*ULd(i)   - FLd(i)   )
                 Rmx = ( aL*ULm(i,1) - FLm(i,1) )
                 Rmy = ( aL*ULm(i,2) - FLm(i,2) )
                 Rmz = ( aL*ULm(i,3) - FLm(i,3) )
                 Re  = ( aL*ULe(i)   - FLe(i)   )
                 Rby = ( aL*ULb(i,2) - FLb(i,2) )
                 Rbz = ( aL*ULb(i,3) - FLb(i,3) )
                 by = f1*Rby
                 bz = f1*Rbz
                 f2 = aL*pt+Re
                 vy = ( (f2-bz*Rbz)*Rmy+by*Rbz*Rmz ) &
                      / ( f2*( f2-by*Rby-bz*Rbz ))
                 vz = ( bz*Rby*Rmy+(f2-by*Rby)*Rmz ) &
                      / ( f2*( f2-by*Rby-bz*Rbz ))
                 vB = vy*by+vz*bz
                 Fd(i)   = f1*Rd*vx
                 Fm(i,1) = aL*( f1*f2*vx ) - Rmx
                 Fm(i,2) = aL*( f1*f2*vy - vB*by ) - Rmy
                 Fm(i,3) = aL*( f1*f2*vz - vB*bz ) - Rmz
                 Fe(i)   = f1*f2*vx
                 Fb(i,1) = 0.d0
                 Fb(i,2) = by*vx
                 Fb(i,3) = bz*vx
!             HLLF flux : F_aR = FR + aR*(U_aR-UR)
!             notice that F_aR != F( U_aR )
              else
                 f1 = 1.d0 / ( aR - vx )
                 Rd  = ( aR*URd(i)   - FRd(i)   )
                 Rmx = ( aR*URm(i,1) - FRm(i,1) )
                 Rmy = ( aR*URm(i,2) - FRm(i,2) )
                 Rmz = ( aR*URm(i,3) - FRm(i,3) )
                 Re  = ( aR*URe(i)   - FRe(i)   )
                 Rby = ( aR*URb(i,2) - FRb(i,2) )
                 Rbz = ( aR*URb(i,3) - FRb(i,3) )
                 by = f1*Rby
                 bz = f1*Rbz
                 f2 = aR*pt+Re
                 vy = ( (f2-bz*Rbz)*Rmy+by*Rbz*Rmz ) &
                      / ( f2*( f2-by*Rby-bz*Rbz ))
                 vz = ( bz*Rby*Rmy+(f2-by*Rby)*Rmz ) &
                      / ( f2*( f2-by*Rby-bz*Rbz ))
                 vB = vy*by+vz*bz
                 Fd(i)   = f1*Rd*vx
                 Fm(i,1) = aR*( f1*f2*vx ) - Rmx
                 Fm(i,2) = aR*( f1*f2*vy - vB*by ) - Rmy
                 Fm(i,3) = aR*( f1*f2*vz - vB*bz ) - Rmz
                 Fe(i)   = f1*f2*vx
                 Fb(i,1) = 0.d0
                 Fb(i,2) = by*vx
                 Fb(i,3) = bz*vx
              endif
           endif
        endif
! ========== HLLD module ==========

     endif
! ========== standard HLL ==========

  enddo

  return
end subroutine hll
