subroutine u2v(Ud,Um,Ue,Ub,Vd,Vv,Vp,ix)
!-----------------------------------------------------------------------
!     Gamma-law version of Mignone & McKinney (2007)'s inversion scheme
!     See also PLUTO (Mignone et al. 2007 ApJS)
!-----------------------------------------------------------------------
!     2009/07/27  S. Zenitani
!-----------------------------------------------------------------------
  implicit none
  integer, parameter :: loop_max = 50        ! Usually 10-20 is OK.
  real(8), parameter :: threshold = 1.0d-13  ! May be computer-dependent
  real(8) :: Ud(ix)    ! density          [input]
  real(8) :: Um(ix,3)  ! momentum density [input]
  real(8) :: Ue(ix)    ! energy density   [input]
  real(8) :: Ub(ix,3)  ! magnetic field   [input]
  real(8) :: Vd(ix)    ! proper density  [output]
  real(8) :: Vv(ix,3)  ! 3-velocity      [output]
  real(8) :: Vp(ix)    ! proper pressure [output]
  integer :: ix
!-----------------------------------------------------------------------
  integer :: i, j, izero
  real(8) :: B2, M2, S, S2, W, W2, W3, S2_W2
  real(8) :: var1, var2, v2, u0, u02
  real(8) :: dv2_dW, chi, dchi_dW, dfW, dW, fW
  real(8) :: ro, pr
      
  do i=1,ix
     B2 = dot_product( Ub(i,1:3), Ub(i,1:3) )
     M2 = dot_product( Um(i,1:3), Um(i,1:3) )
     S  = dot_product( Ub(i,1:3), Um(i,1:3) )
     S2 = S*S
     
     var1 = - 4 * ( Ue(i) - B2 )
     var2 = M2 + ( B2 - 2*Ue(i) )*B2
     W = ( - var1 + sqrt(var1**2 - 12*var2))/6.d0
     if( W.lt.Ud(i) ) then
        W = Ud(i)
     elseif( W.gt.(4.d0/3.d0)*(Ue(i)-B2/2.d0) ) then
        W = (4.d0/3.d0)*(Ue(i)-B2/2.d0)
     endif
     
! ---- iteration loop ----
     do j=1,loop_max

        W2 = W*W
        W3 = W2*W
        S2_W2 = S2/W2
        var1  = W + B2

        v2 = (M2 + S2_W2*(var1 + W))/(var1*var1)
        if( v2.gt.1 )then
           write(6,*) 'stop: v2 > 1', j, 'th iteration'
           stop
        endif

        u02 = 1.d0 / (1.d0 - v2)
        u0  = sqrt(u02)
        
        dv2_dW  = ( S2*(3*W*var1 + B2*B2) + M2*W3 ) * (-2) / W3 / (var1**3)
        chi = (W - Ud(i)*u0) * (1.d0 - v2)
        ro  = Ud(i)/u0
        dchi_dW =  1.d0 - v2 - 0.5d0*u0*(Ud(i) + 2*chi*u0)*dv2_dW
        pr = 0.25d0*chi

        fW  = W - pr + 0.5d0*(1.d0 + v2)*B2 - 0.5d0*S2_W2 - Ue(i)
        dfW = 1.d0 - 0.25d0*dchi_dW + 0.5d0*B2*dv2_dW + S2/W3
        dW = fW/dfW

        W = W - dW
!        write(6,*) j, ' W, dW, dW/W', W, dW, dW/W
        if (abs(dW).lt.(threshold*W)) then
           if( pr.gt.0.d0 ) then
              goto 1000
           endif
        endif
        if( j.eq.loop_max ) then
           write(6,*) 'No convergence'
           write(6,*) 'U: ', i, Ud(i), Ue(i), Um(i,1), Um(i,2), Um(i,3)
           stop
        endif

     enddo
1000 continue
! ---- iteration loop ----

!     write(6,*) 'error: ', abs(dW)/W

     var1 = 1.d0/(W + B2)
     var2 = S/W

     Vp(i)   = pr
     Vv(i,1:3) = var1*( Um(i,1:3) + var2*Ub(i,1:3) )
     v2 = dot_product( Vv(i,1:3), Vv(i,1:3) )
     if( v2 .ge. 1 ) then
        write(6,*) 'error: superluminous velocity', i, v2
        stop
     endif
     Vd(i) = Ud(i)*sqrt( 1.d0 - v2 )
     
  enddo

  return
end subroutine u2v
